"""Beam - MySQL error classes."""


class BeamMySQLError(Exception):
    """Base class for all Beam MySQL errors."""


class MySQLClientError(BeamMySQLError):
    """An error in the mysql client object (e.g. failed to execute query)."""



"""I/O connectors of mysql."""

from typing import Dict
from typing import Union

import apache_beam as beam
from apache_beam.io import iobase
from apache_beam.options.value_provider import ValueProvider
from apache_beam.pvalue import PCollection
from apache_beam.transforms.core import PTransform

from beam_mysql.connector import splitters
from beam_mysql.connector.client import MySQLClient
from beam_mysql.connector.source import MySQLSource
from beam_mysql.connector.utils import get_runtime_value


class ReadFromMySQL(PTransform):
    """Create PCollection from MySQL."""

    def __init__(
        self,
        query: Union[str, ValueProvider],
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider] = 3306,
        splitter=splitters.NoSplitter(),
    ):
        super().__init__()
        self._query = query
        self._host = host
        self._database = database
        self._user = user
        self._password = password
        self._port = port
        self._splitter = splitter

    def expand(self, pcoll: PCollection) -> PCollection:
        return pcoll | iobase.Read(
            MySQLSource(self._query, self._host, self._database, self._user, self._password, self._port, self._splitter)
        )


class WriteToMySQL(PTransform):
    """Write dict rows to MySQL."""

    def __init__(
        self,
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        table: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider] = 3306,
        batch_size: int = 1000,
    ):
        super().__init__()
        self._host = host
        self._database = database
        self._table = table
        self._user = user
        self._password = password
        self._port = port
        self._batch_size = batch_size

    def expand(self, pcoll: PCollection) -> PCollection:
        return pcoll | beam.ParDo(
            _WriteToMySQLFn(
                self._host, self._database, self._table, self._user, self._password, self._port, self._batch_size
            )
        )


class _WriteToMySQLFn(beam.DoFn):
    """DoFn for WriteToMySQL."""

    def __init__(
        self,
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        table: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider],
        batch_size: int,
    ):
        super().__init__()
        self._host = host
        self._database = database
        self._table = table
        self._user = user
        self._password = password
        self._port = port
        self._batch_size = batch_size

        self._config = {
            "host": self._host,
            "database": self._database,
            "user": self._user,
            "password": self._password,
            "port": self._port,
        }

    def start_bundle(self):
        self._build_value()
        self._queries = []

    def process(self, element: Dict, *args, **kwargs):
        columns = []
        values = []
        for column, value in element.items():
            columns.append(column)
            values.append(value)

        column_str = ", ".join(columns)
        value_str = ", ".join([f"{value}" if isinstance(value, (int, float)) else f"'{value}'" for value in values])
        query = f"INSERT INTO {self._config['database']}.{self._table}({column_str}) VALUES({value_str});"

        self._queries.append(query)

        if len(self._queries) > self._batch_size:
            self._client.record_loader("\n".join(self._queries))
            self._queries.clear()

    def finish_bundle(self):
        if len(self._queries):
            self._client.record_loader("\n".join(self._queries))
            self._queries.clear()

    def _build_value(self):
        for k, v in self._config.items():
            self._config[k] = get_runtime_value(v)
        self._table = get_runtime_value(self._table)
        self._batch_size = get_runtime_value(self._batch_size)

        self._client = MySQLClient(self._config)



"""Utility functions."""

from typing import Any

from apache_beam.options.value_provider import ValueProvider


def get_runtime_value(value: Any) -> Any:
    if isinstance(value, ValueProvider):
        return value.get()
    else:
        return value


def cleanse_query(query: str) -> str:
    return query.strip(";")



"""A source that reads a finite amount of records on mysql."""

from typing import Union

from apache_beam.io import iobase
from apache_beam.options.value_provider import ValueProvider

from beam_mysql.connector import splitters
from beam_mysql.connector.client import MySQLClient
from beam_mysql.connector.utils import cleanse_query
from beam_mysql.connector.utils import get_runtime_value


class MySQLSource(iobase.BoundedSource):
    """A source object of mysql."""

    def __init__(
        self,
        query: Union[str, ValueProvider],
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider],
        splitter: splitters.BaseSplitter,
    ):
        super().__init__()
        self._query = query
        self._host = host
        self._database = database
        self._user = user
        self._password = password
        self._port = port

        self._is_builded = False

        self._config = {
            "host": self._host,
            "database": self._database,
            "user": self._user,
            "password": self._password,
            "port": self._port,
        }

        self._splitter = splitter

    def estimate_size(self):
        """Implement :class:`~apache_beam.io.iobase.BoundedSource.estimate_size`"""
        if not self._is_builded:
            self._build_value()

        return self._splitter.estimate_size()

    def get_range_tracker(self, start_position, stop_position):
        """Implement :class:`~apache_beam.io.iobase.BoundedSource.get_range_tracker`"""
        if not self._is_builded:
            self._build_value()

        return self._splitter.get_range_tracker(start_position, stop_position)

    def read(self, range_tracker):
        """Implement :class:`~apache_beam.io.iobase.BoundedSource.read`"""
        for record in self._splitter.read(range_tracker):
            yield record

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        """Implement :class:`~apache_beam.io.iobase.BoundedSource.split`"""
        if not self._is_builded:
            self._build_value()

        for split in self._splitter.split(desired_bundle_size, start_position, stop_position):
            yield split

    def _build_value(self):
        for k, v in self._config.items():
            self._config[k] = get_runtime_value(v)

        self.query = cleanse_query(get_runtime_value(self._query))
        self.client = MySQLClient(self._config)
        self._splitter.build_source(self)

        self._is_builded = True



"""A client of mysql."""

import logging
from typing import Dict
from typing import Generator
from typing import List

import mysql.connector
from mysql.connector.errors import Error as MySQLConnectorError

from beam_mysql.connector.errors import MySQLClientError

_SELECT_STATEMENT = "SELECT"
_INSERT_STATEMENT = "INSERT"


class MySQLClient:
    """A mysql client object."""

    def __init__(self, config: Dict):
        self._config = config
        self._validate_config(self._config)

    def record_generator(self, query: str, dictionary=True) -> Generator[Dict, None, None]:
        """
        Generate dict record from raw data on mysql.

        Args:
            query: query with select statement
            dictionary: the type of result is dict if true else tuple

        Returns:
            dict record

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, [_SELECT_STATEMENT])

        with _MySQLConnection(self._config) as conn:
            # buffered is false because it can be assumed that the data size is too large
            cur = conn.cursor(buffered=False, dictionary=dictionary)

            try:
                cur.execute(query)
                logging.info(f"Successfully execute query: {query}")

                for record in cur:
                    yield record
            except MySQLConnectorError as e:
                raise MySQLClientError(f"Failed to execute query: {query}, Raise exception: {e}")

            cur.close()

    def counts_estimator(self, query: str) -> int:
        """
        Make a estimate of the total number of records.

        Args:
            query: query with select statement

        Returns:
            the total number of records

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, [_SELECT_STATEMENT])
        count_query = f"SELECT COUNT(*) AS count FROM ({query}) as subq"

        with _MySQLConnection(self._config) as conn:
            # buffered is false because it can be assumed that the data size is too large
            cur = conn.cursor(buffered=False, dictionary=True)

            try:
                cur.execute(count_query)
                logging.info(f"Successfully execute query: {count_query}")

                record = cur.fetchone()
            except MySQLConnectorError as e:
                raise MySQLClientError(f"Failed to execute query: {count_query}, Raise exception: {e}")

            cur.close()

            return record["count"]

    def rough_counts_estimator(self, query: str) -> int:
        """
        Make a rough estimate of the total number of records.
        To avoid waiting time by select counts query when the data size is too large.

        Args:
            query: query with select statement

        Returns:
            the total number of records

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, [_SELECT_STATEMENT])
        count_query = f"EXPLAIN SELECT * FROM ({query}) as subq"

        with _MySQLConnection(self._config) as conn:
            # buffered is false because it can be assumed that the data size is too large
            cur = conn.cursor(buffered=False, dictionary=True)

            try:
                cur.execute(count_query)
                logging.info(f"Successfully execute query: {count_query}")

                records = cur.fetchall()

                total_number = 0

                for record in records:
                    # Query of the argument should be "DERIVED" because it is sub query of explain select.
                    # Count query should be "PRIMARY" or "SIMPLE" because it is not sub query.
                    if record["select_type"] in ("PRIMARY", "SIMPLE"):
                        total_number = record["rows"]
            except MySQLConnectorError as e:
                raise MySQLClientError(f"Failed to execute query: {count_query}, Raise exception: {e}")

            cur.close()

            if total_number <= 0:
                raise mysql.connector.errors.Error(f"Failed to estimate total number of records. Query: {count_query}")
            else:
                return total_number

    def record_loader(self, query: str):
        """
        Load dict record into mysql.

        Args:
            query: query with insert or update statement

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, [_INSERT_STATEMENT])

        with _MySQLConnection(self._config) as conn:
            cur = conn.cursor()

            try:
                cur.execute(query)
                conn.commit()
                logging.info(f"Successfully execute query: {query}")
            except MySQLConnectorError as e:
                conn.rollback()
                raise MySQLClientError(f"Failed to execute query: {query}, Raise exception: {e}")

            cur.close()

    @staticmethod
    def _validate_config(config: Dict):
        required_keys = {"host", "port", "database", "user", "password"}
        if not config.keys() == required_keys:
            raise MySQLClientError(f"Config is not satisfied. required: {required_keys}, actual: {config.keys()}")

    @staticmethod
    def _validate_query(query: str, statements: List[str]):
        query = query.lstrip()

        for statement in statements:
            if statement and not query.lower().startswith(statement.lower()):
                raise MySQLClientError(f"Query expected to start with {statement} statement. Query: {query}")


class _MySQLConnection:
    """A wrapper object to connect mysql."""

    def __init__(self, _config: Dict):
        self._config = _config

    def __enter__(self):
        try:
            self.conn = mysql.connector.connect(**self._config)
            return self.conn
        except MySQLConnectorError as e:
            raise MySQLClientError(f"Failed to connect mysql, Raise exception: {e}")

    def __exit__(self, exception_type, exception_value, traceback):
        self.conn.close()



"""Beam - MySQL Connector for setup file."""

import os
import subprocess

from setuptools import Command
from setuptools import find_packages
from setuptools import setup

SRC_DIR_NAME = "beam_mysql"
EXAMPLES_DIR_NAME = "examples"
TESTS_DIR_NAME = "tests"


class SimpleCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class LintCommand(SimpleCommand):
    def run(self):
        subprocess.run(["black", SRC_DIR_NAME, EXAMPLES_DIR_NAME, TESTS_DIR_NAME, "-l", "120", "--check", "--diff"])
        subprocess.run(["mypy", SRC_DIR_NAME, EXAMPLES_DIR_NAME, TESTS_DIR_NAME])


class FormatCommand(SimpleCommand):
    def run(self):
        subprocess.run(["black", SRC_DIR_NAME, EXAMPLES_DIR_NAME, TESTS_DIR_NAME, "-l", "120"])


def get_version():
    global_names = {}
    exec(open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "beam_mysql/__init__.py")).read(), global_names)
    return global_names["__version__"]


with open("README.md") as f:
    README = f.read()


PACKAGE_NAME = "beam-mysql-connector"
PACKAGE_VERSION = get_version()
PACKAGE_DESCRIPTION = "MySQL I/O Connector of Apache Beam"
PACKAGE_URL = "https://github.com/esaki01/beam-mysql-connector"
PACKAGE_DOWNLOAD_URL = "https://pypi.python.org/pypi/beam-mysql-connector"
PACKAGE_AUTHOR = "esaki01"
PACKAGE_EMAIL = "esaki1011@gmail.com"
PACKAGE_KEYWORDS = "apache beam mysql connector"
PACKAGE_LONG_DESCRIPTION = README

REQUIRED_PACKAGES = [
    "apache-beam>=2.15.*",
    "mysql-connector-python>=8.0.*",
]

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description=PACKAGE_DESCRIPTION,
    long_description=PACKAGE_LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=PACKAGE_URL,
    download_url=PACKAGE_DOWNLOAD_URL,
    author=PACKAGE_AUTHOR,
    author_email=PACKAGE_EMAIL,
    packages=find_packages(),
    install_requires=REQUIRED_PACKAGES,
    python_requires=">=3.5",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    license="MIT",
    keywords=PACKAGE_KEYWORDS,
    cmdclass=dict(lint=LintCommand, fmt=FormatCommand),
)



"""A wrapper class of `~apache_beam.io.iobase.BoundedSource` functions."""

import re
from abc import ABCMeta
from abc import abstractmethod
from datetime import datetime
from typing import Callable
from typing import Iterator

from apache_beam.io import iobase
from apache_beam.io.range_trackers import LexicographicKeyRangeTracker
from apache_beam.io.range_trackers import OffsetRangeTracker
from apache_beam.io.range_trackers import UnsplittableRangeTracker
from dateutil.relativedelta import relativedelta


class BaseSplitter(metaclass=ABCMeta):
    """Abstract class of splitter."""

    def build_source(self, source):
        """Build source on runtime."""
        self.source = source

    @abstractmethod
    def estimate_size(self):
        """Wrap :class:`~apache_beam.io.iobase.BoundedSource.estimate_size`"""
        raise NotImplementedError()

    @abstractmethod
    def get_range_tracker(self, start_position, stop_position):
        """Wrap :class:`~apache_beam.io.iobase.BoundedSource.get_range_tracker`"""
        raise NotImplementedError()

    @abstractmethod
    def read(self, range_tracker):
        """Wrap :class:`~apache_beam.io.iobase.BoundedSource.read`"""
        raise NotImplementedError()

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        """Wrap :class:`~apache_beam.io.iobase.BoundedSource.split`"""
        raise NotImplementedError()


class NoSplitter(BaseSplitter):
    """No split bounded source so not work parallel."""

    def estimate_size(self):
        return self.source.client.rough_counts_estimator(self.source.query)

    def get_range_tracker(self, start_position, stop_position):
        if start_position is None:
            start_position = 0
        if stop_position is None:
            stop_position = OffsetRangeTracker.OFFSET_INFINITY

        return UnsplittableRangeTracker(OffsetRangeTracker(start_position, stop_position))

    def read(self, range_tracker):
        for record in self.source.client.record_generator(self.source.query):
            yield record

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        if start_position is None:
            start_position = 0
        if stop_position is None:
            stop_position = OffsetRangeTracker.OFFSET_INFINITY

        yield iobase.SourceBundle(
            weight=desired_bundle_size, source=self.source, start_position=start_position, stop_position=stop_position
        )


class LimitOffsetSplitter(BaseSplitter):
    """Split bounded source by limit and offset."""

    def __init__(self, batch_size: int = 1000000):
        self._batch_size = batch_size
        self._counts = 0

    def estimate_size(self):
        self._counts = self.source.client.counts_estimator(self.source.query)
        return self._counts

    def get_range_tracker(self, start_position, stop_position):
        if self._counts == 0:
            self._counts = self.source.client.counts_estimator(self.source.query)
        if start_position is None:
            start_position = 0
        if stop_position is None:
            stop_position = self._counts

        return LexicographicKeyRangeTracker(start_position, stop_position)

    def read(self, range_tracker):
        offset, limit = range_tracker.start_position(), range_tracker.stop_position()
        query = f"SELECT * FROM ({self.source.query}) as subq LIMIT {limit} OFFSET {offset}"
        for record in self.source.client.record_generator(query):
            yield record

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        if self._counts == 0:
            self._counts = self.source.client.counts_estimator(self.source.query)
        if start_position is None:
            start_position = 0
        if stop_position is None:
            stop_position = self._counts

        last_position = 0
        for offset in range(start_position, stop_position, self._batch_size):
            yield iobase.SourceBundle(
                weight=desired_bundle_size, source=self.source, start_position=offset, stop_position=self._batch_size
            )
            last_position = offset + self._batch_size

        yield iobase.SourceBundle(
            weight=desired_bundle_size,
            source=self.source,
            start_position=last_position + 1,
            stop_position=stop_position,
        )


class IdsSplitter(BaseSplitter):
    """Split bounded source by any ids."""

    def __init__(self, generate_ids_fn: Callable[[], Iterator], batch_size: int = 1000000):
        self._generate_ids_fn = generate_ids_fn
        self._batch_size = batch_size

    def estimate_size(self):
        # TODO: unify the method of estimating
        return 0

    def get_range_tracker(self, start_position, stop_position):
        self._validate_query()
        return LexicographicKeyRangeTracker(start_position, stop_position)

    def read(self, range_tracker):
        if range_tracker.start_position() is None:
            ids = ",".join([f"'{id}'" for id in self._generate_ids_fn()])
        else:
            ids = range_tracker.start_position()

        query = self.source.query.format(ids=ids)
        for record in self.source.client.record_generator(query):
            yield record

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        self._validate_query()

        ids = []
        for generated_id in self._generate_ids_fn():
            ids.append(generated_id)
            if len(ids) == self._batch_size:
                yield self._create_bundle_source(desired_bundle_size, self.source, ids)
                ids.clear()
        yield self._create_bundle_source(desired_bundle_size, self.source, ids)

    def _validate_query(self):
        condensed_query = self.source.query.lower().replace(" ", "")
        if re.search(r"notin\({ids}\)", condensed_query):
            raise ValueError(f"Not support 'not in' phrase: {self.source.query}")
        if not re.search(r"in\({ids}\)", condensed_query):
            example = "SELECT * FROM tests WHERE id IN ({ids})"
            raise ValueError(f"Require 'in' phrase and 'ids' key on query: {self.source.query}, e.g. '{example}'")

    @staticmethod
    def _create_bundle_source(desired_bundle_size, source, ids):
        if isinstance(ids, list):
            ids_str = ",".join([f"'{id}'" for id in ids])
        elif isinstance(ids, str):
            ids_str = ids
        else:
            raise ValueError(f"Unexpected ids: {ids}")

        return iobase.SourceBundle(
            weight=desired_bundle_size, source=source, start_position=ids_str, stop_position=None
        )


class PartitionSplitter(BaseSplitter):
    """Split bounded source by partitions."""

    PATTERN = r".*\((,?p\d{6})+\).*"

    def estimate_size(self):
        return self.source.client.rough_counts_estimator(self.source.query)

    def get_range_tracker(self, start_position, stop_position):
        self._validate_query()
        return LexicographicKeyRangeTracker(start_position, stop_position)

    def read(self, range_tracker):
        if range_tracker.start_position() is None:
            query = self.source.query
        else:
            partition, partitions = range_tracker.start_position(), range_tracker.stop_position()
            query = self.source.query.replace(partitions, partition)

        for record in self.source.client.record_generator(query):
            yield record

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        self._validate_query()

        query = self.source.query
        partitions = []
        while True:
            match = re.match(self.PATTERN, query)
            if not match:
                break

            partition = match.group(1)
            query = query.replace(partition, "")
            partitions.append(partition)

        partitions.reverse()
        for p in partitions:
            partition = p.replace(",", "")
            yield iobase.SourceBundle(
                weight=desired_bundle_size,
                source=self.source,
                start_position=partition,
                stop_position="".join(partitions),
            )

    def _validate_query(self):
        if not re.search(self.PATTERN, self.source.query):
            example = "SELECT * FROM tests PARTITION (p202001,p202002)"
            raise ValueError(f"Require 'partition' phrase on query: {self.source.query}, e.g. '{example}'")


class DateSplitter(BaseSplitter):
    """Split bounded source by dates."""

    PATTERN = r".*[\'\"]*(\d{4}-\d{2}-\d{2})[\'\"]*[\w\s]+[\'\"]*(\d{4}-\d{2}-\d{2})[\'\"]*.*"

    def estimate_size(self):
        return self.source.client.rough_counts_estimator(self.source.query)

    def get_range_tracker(self, start_position, stop_position):
        self._validate_query()
        return LexicographicKeyRangeTracker(start_position, stop_position)

    def read(self, range_tracker):
        if range_tracker.start_position() is None:
            query = self.source.query
        else:
            start_date, end_date = range_tracker.start_position(), range_tracker.stop_position()
            match = re.match(self.PATTERN, self.source.query)
            query = self.source.query.replace(match.group(1), start_date).replace(match.group(2), end_date)

        for record in self.source.client.record_generator(query):
            yield record

    def split(self, desired_bundle_size, start_position=None, stop_position=None):
        self._validate_query()

        match = re.match(self.PATTERN, self.source.query)
        start_date = datetime.strptime(match.group(1), "%Y-%m-%d")
        end_date = datetime.strptime(match.group(2), "%Y-%m-%d")

        months = self._diff_between_dates(start_date, end_date)
        for month in months:
            yield iobase.SourceBundle(
                weight=desired_bundle_size, source=self.source, start_position=month[0], stop_position=month[1],
            )

    def _validate_query(self):
        if not re.search(self.PATTERN, self.source.query):
            example = "SELECT * FROM tests WHERE date BETWEEN 2019-01-01 AND 2020-01-01"
            raise ValueError(f"Require 'between' phrase on query: {self.source.query}, e.g. '{example}'")

    @staticmethod
    def _diff_between_dates(start_date, end_date):
        diff_months = (end_date.year * 12 + end_date.month) - (start_date.year * 12 + start_date.month)
        tuple_months = [
            (start_date + relativedelta(months=i), start_date + relativedelta(months=i + 1) - relativedelta(days=1))
            for i in range(diff_months + 1)
        ]
        last_date = tuple_months[-1][1]
        tuple_months[-1] = (tuple_months[-1][0], datetime(last_date.year, last_date.month, end_date.day))

        return [
            (tuple_month[0].strftime("%Y-%m-%d"), tuple_month[1].strftime("%Y-%m-%d")) for tuple_month in tuple_months
        ]
